/* Copyright (c) 2009 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  tcjs.h - tads 3 compiler - javascript code generator
Function
  
Notes
  
Modified
  02/17/09 MJRoberts  - Creation
*/

#ifndef TCJS_H
#define TCJS_H

/* include the Javascript code generator final parse node classes */
#include "tcjsdrv.h"

/* ------------------------------------------------------------------------ */
/*
 *   T3-specific code generator helper class.  This class provides a set
 *   of static functions that are useful for T3 code generation.  
 */
class CTcGenTarg
{
public:
    /* initialize the code generator */
    CTcGenTarg();

    /* destroy the code generator object */
    ~CTcGenTarg();

    /* 
     *   Push a javascript expression.  The string can be a printf-style
     *   format string, with varargs following.  First, the printf formatting
     *   is applied to generate the control string.  Then, we expand '$'
     *   expressions:
     *   
     *   $1 = the top-of-stack code expression
     *.  $2 = the 2nd item on the stack
     *.  $n = the nth item on the stack
     *   
     *   ${disc n} = substitute an empty string, but mark the nth stack item
     *   for discarding at the end of the process (see below) as though it
     *   had been mentioned in a regular $n expression
     *   
     *   ${args n m} = the 'm' stack elements starting at $n, concatenated
     *   together into an argument list (with commas as separators)
     *   
     *   $$ = a single literal '$'
     *   
     *   After expanding the '$' expressions, we pop the stack expressions
     *   used.  We pop every stack item up to and including the highest
     *   numbered item used in a '$' expression - even if an item is never
     *   used, it will be discarded if it's on the stack after an item that
     *   was mentioned.
     *   
     *   Finally, the expanded expression string that results from the above
     *   substitutions is pushed onto the expression stack.  
     */
    void js_expr(const char *str, ...);

    /* allocate a new global property ID */
    tctarg_prop_id_t new_prop_id() { return next_prop_++; }

    /* allocate a new global object ID */
    tctarg_obj_id_t new_obj_id() { return next_obj_++; }

    /* 
     *   Note constant data sizes.  This doesn't affect us, since we don't
     *   have to worry about dividing anything up into pages the way the T3
     *   code generator does.  
     */
    void note_str(size_t) { }
    void note_list(size_t) { }

    /* get the maximum string/list/bytecode lengths */
    size_t get_max_str_len() const { return 0; }
    size_t get_max_list_cnt() const { return 0; }
    size_t get_max_bytecode_len() const { return 0; }

    /* 
     *   Add a debug line record.  We don't keep debugging information, so we
     *   can just ignore this. 
     */
    void add_line_rec(class CTcTokFileDesc *, long) { }

    /*
     *   Add a function set to the dependency table - returns the index of
     *   the function set in the table 
     */
    int add_fnset(const char *fnset_extern_name, size_t len);
    int add_fnset(const char *fnset_extern_name)
        { return add_fnset(fnset_extern_name, strlen(fnset_extern_name)); }

    /* get the name of a function set given its index */
    const char *get_fnset_name(int idx) const;

    /* get the number of defined function sets */
    int get_fnset_cnt() const { return fnset_cnt_; }

    /* 
     *   Find a metaclass entry, adding it if it's not already there.  If the
     *   metaclass is already defined, and it has an associated symbol, we
     *   will not change the associated symbol - this will let the caller
     *   detect that the metaclass has been previously defined for a
     *   different symbol, which is usually an error.  
     */
    int find_or_add_meta(const char *nm, size_t len,
                         class CTcSymMetaclass *sym);

    /* get/set the symbol for a given metaclass */
    class CTcSymMetaclass *get_meta_sym(int meta_idx);
    void set_meta_sym(int meta_idx, class CTcSymMetaclass *sym);

    /* get the number of metaclasses */
    int get_meta_cnt() const { return meta_cnt_; }

    /* get the external (universally unique) name for the given metaclass */
    const char *get_meta_name(int idx) const;

    /*
     *   Notify the code generator that we're replacing an object (via the
     *   "replace" statement) at the given stream offset.  We'll mark the
     *   data in the stream as deleted so that we don't write it to the image
     *   file.  
     */
    void notify_replace_object(ulong stream_ofs);

    /*
     *   Notify the code generator that parsing is finished.  This should be
     *   called after parsing and before code generation begins.  
     */
    void parsing_done();

    /* generate synthesized code during linking */
    void build_synthesized_code();

    /* generate code for a dictionary object */
    void gen_code_for_dict(class CTcDictEntry *dict);

    /* generate code for a grammar production object */
    void gen_code_for_gramprod(class CTcGramProdEntry *prod);

    /*
     *   Write to an object file.  The compiler calls this after all parsing
     *   and code generation are completed to write an object file, which can
     *   then be linked with other object files to create an image file.  
     */
    void write_to_object_file(class CVmFile *object_fp,
                              class CTcMake *make_obj);

    /*
     *   Load an object file.  Returns zero on success, non-zero on error.  
     */
    int load_object_file(CVmFile *fp, const textchar_t *fname);

    /*
     *   Write the image file.  The compiler calls this after all parsing and
     *   code generation are completed to write an image file.  We must apply
     *   all fixups, assign the code and constant pool layouts, and write the
     *   data to the image file.  
     */
    void write_to_image(class CVmFile *image_fp, uchar data_xor_mask,
                        const char tool_data[4]);

protected:
    /* property/object ID allocators */
    tctarg_prop_id_t next_prop_;
    tctarg_obj_id_t next_obj_;

    /* metaclass tracking */
    int meta_cnt_;

    /* function set tracking */
    int fnset_cnt_;
};


#endif /* TCJS_H */
